#!/usr/bin/env bash
#
# Copyright(c) 2006-2017 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#
#	sudo ipsetapply.sh TS.org
#

#######################################################
################## FUNCTIONS PART #####################
#######################################################

#######################################
function verifyRuleExistence() {
#XXX! we only use old ipset syntax here for now because ipsetget generates only old syntax
#when generating the file we pipe in to ipset and this is easier to support across all vyatta versions.
	local rulename=$1
	local ipsetsize=$2
	local ruleOk=0

	if ! $ipset -S | grep -E " ${tsprefix}${rulename} " 1>/dev/null 2>/dev/null; then
		echo "[INFO ] : Missing ruleset [${tsprefix}${rulename}]... Creating it ..."
		$ipset -N  ${tsprefix}${rulename} nethash --hashsize 32768 --maxelem $ipsetsize
	fi

	return 0
}

#######################################################
################### MAIN PART #########################
#######################################################
echo
echo "[INFO ] : Starting ipsetapply operation."

# Check if running as root. If not sudo and rerun
echo "[INFO ] : Verifying running UID ..."
if [ `whoami` != root ]; then
	echo "[ERROR] : Need to be run as root... Exiting ..."
	#sudo $0 $1
	exit 1
fi

# Change the path to the config file if it is no located in /usr/local/etc
CONFIG="threatstop.conf"

# Check for and read the config file, abort if not found
echo "[INFO ] : Loading configuration data."
if [ -f $CONFIG ]; then
	. $CONFIG
	TSNEW="${1:-TS.new}";
elif [ -f "`dirname $0`/$CONFIG" ]; then
	. "`dirname $0`/$CONFIG"
	TSNEW="`dirname $0`/${1:-TS.new}"
	echo "Using `dirname $0`/$CONFIG.  Applying $TSNEW for $ipset" >>/tmp/tsoutput.txt
else
	echo "[CRITICAL] : Cannot read config file: \"$CONFIG\" or \"`dirname $0`/$CONFIG\"... Exiting."
	exit 1
fi

echo "[INFO ] : Verifying Block lists existence."
BLIST=`sed  '/[0-9]\{1,3\}\.[0-9]\{1,3\}\.[0-9]\{1,3\}\.[0-9]\{1,3\}/!d' $TSNEW | wc -l`

if [ $BLIST  -lt 4 ]; then
	echo "ERROR : No data received from DNS server."
	exit 1
fi

# verify ruleset existence and recreate them if needed
echo "[INFO ] : Verifying ruleset existence."
verifyRuleExistence blocknet $block_ipsetsize
verifyRuleExistence allownet $allow_ipsetsize

echo "[INFO ] : Performing ipset commands."
$ipset -R <$TSNEW >>/tmp/tsoutput.txt
$ipset -W ${tsprefix}blocknet ${tsprefix}blocknetnew >>/tmp/tsoutput.txt
$ipset -W ${tsprefix}allownet ${tsprefix}allownetnew >>/tmp/tsoutput.txt
$ipset -X ${tsprefix}blocknetnew >>/tmp/tsoutput.txt
$ipset -X ${tsprefix}allownetnew >>/tmp/tsoutput.txt

echo "ipset done" >>/tmp/tsoutput.txt
echo "[INFO ] : Operation completed ..."
echo
