#!/usr/bin/perl
#
# Copyright(c) 2006-2017 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#
# ThreatSTOP main function module
################################################################################

package ThreatStop;
use Exporter;

use strict;
use warnings;
use Scalar::Util qw(openhandle);
use Cwd qw(abs_path);
use Fcntl qw(:flock);
use File::Basename;
use FindBin;
use Switch;
use Data::Dumper;
use MIME::Base64 qw(encode_base64 decode_base64);
FindBin->again;
use POSIX qw(strftime);

###############################################################
################## MODULE DEFENITIONS PART ####################
###############################################################
our $VERSION   = 1.00;
our @ISA       = qw(Exporter);
our @EXPORT    = qw(getCurrentTime logInit printMsg logMsg setLock setUnLock ipv4_cidr2msk checkDNSaccess calculateElapsedTime loadConfig intHandler osPlatformDetecter loadConfFileData setWorkingDir setAlarm osDestributionDetector appendUploadLogEntry purgeOldUploadsLogs $DNS_LIST $DEBUG_LEVEL $LOGFILE $LOG_LEVEL $dns_server $port);

our $CFG_DATA;
our $CONFIGFILE  = $FindBin::Bin . "/threatstop.conf" if ( ! defined ($CONFIGFILE));
our $LOGFILE     = $FindBin::Bin . "/tsoutput.log" if ( ! defined $LOGFILE); #Define default log
our $UPLOADSLOGFILE = $FindBin::Bin . "/tsuploads";
our $LOCKFILE    = $0.".lck" if ( ! defined ($LOCKFILE));
our $LOGINIT     = 0;
our $DEBUG_LEVEL = 5 if ( ! defined $DEBUG_LEVEL);                        #Define verbose level 5-nothing
our $LOG_LEVEL   = 6 if ( ! defined $LOG_LEVEL);                          #Define verbose level 5-nothing
our ($FD, $FDL, $LFD);
our $START_EPOCH = time();
our $END_EPOCH;

our $DNS_LIST = "192.124.129.42";
our $dns_server;                                                          # "192.124.129.42";
our $DNS_PROTO       = '+tcp';					          # Protocol to use - TCP
our $DNS_PROTO_ALT   = '+notcp';                                          # Protocol to use - UDP
our $DNS_PORT_BASE   = 53;                                                # Port to use (default 53)
our $DNS_PORT_ALT    = 5353;                                              # Port to use (alternate 5353)
our $DNS_DOMAIN_BASE = 'dns-servers.threatstop.local';                    # Server name to resolve 
our $DNS_DOMAIN_ALT  = 'dns-servers.threatstop.com';                      # Alternative server name to resolve
our $DNS_SERVERS_ALT = "8.8.8.8 8.8.4.4 208.67.220.220 208.67.222.222";   # Alternate servers ip's to quiery
our $DNS_SERVERS_LCL = '127.0.0.1';                                       # Local DNS serves will be stored here
our $DNS_LIST_NEW;
our $DNS_PORT        = 53;                                                # Actually we read this param from conf file
our $curr_date = strftime "%m%Y", localtime;

# Setup signal handler for possible user interupts
$SIG{'INT'}  = 'intHandler';
$SIG{'ABRT'} = 'intHandler';
$SIG{'QUIT'} = 'intHandler';
$SIG{'STOP'} = 'intHandler';

###############################################################
################# FUNCTIONS DEFENITIONS PART ##################
###############################################################

############################################
########## PRIVATE FUNCTIONS ###############
############################################

################################################################

##################################################
# intHandler : Used to trap exit signals and per #
#              form lockfile cleanup             #
##################################################
sub intHandler() {
        if ( defined $FDL ) {
                close($FDL);
                unlink($LOCKFILE);
        }

        print "\n";
        exit(0);
}

##################################################
# loadConfig : Loads additional parameters from  #
#              conf file                         #
##################################################
sub loadConfig() {

        &printMsg("INFO","Loading configuration file ...");

        eval {
                open(FD,"<$CONFIGFILE") or die "Failed to open configuration file [$CONFIGFILE].  The following error occurred [$!]";
        };

        if ( $@ ) {
                &printMsg("ERROR","$@");
                return 1;
        }

        while (<FD>) {
                chomp;                  # no newline
                s/#.*//;                # no comments
                s/^\s+//;               # no leading white
                s/\s+$//;               # no trailing white
                s/'//g;                 # no ticks
                next unless length;     # anything left?
                my ($var, $value) = split(/\s*=\s*/, $_, 2);
                $CFG_DATA->{$var} = $value;
        }

	close FD;
    mpw($CONFIGFILE,$CFG_DATA);
	return 0;
}

#mangle/unmangle config must be called right after loading in config values. 
sub mpw {
   my $conf = shift; 
   my $confref = shift;
   foreach my $i (qw(password enable_pw)) {
       if(defined($confref->{$i})) {
           my $tmp = $confref->{$i};
           $tmp =~ s/\"//g;
           if($tmp =~ /^\$\$.*\$\$$/) {
               $tmp =~ s/\$//g;
               $tmp =~ s/!/=/g;
               $tmp = decode_base64($tmp);
#               print "$i encoded\n";
#               print "$i: $tmp\n";
               $confref->{$i} = $tmp;
           }
           else {
#               print "$i plaintext\n";
#               print "$i: $confref->{$i}\n";
               my $tmp = $confref->{$i};
               $tmp =~ s/\"//g;
               $tmp = encode_base64($tmp);
               chomp($tmp);
               $tmp =~ s/=/!/g;
               $tmp = '"' . '$$' . $tmp . '$$' . '"';
               rewriteConfKey($conf,$i,$tmp);
           }
           
       }
   }
}

#Re-write the given key/value pair in the specified config file.
sub rewriteConfKey() {
   my $conf = shift;
   my $key = shift;
   my $value = shift;
   my $in;
   my $out;
   my @data;
   eval {
       open($in,'<',$conf);
       @data = <$in>;
       close($in);
       open($out,'>',$conf);
   };
   if ( $@ ) {
       &printMsg("ERROR",$@);
       return 1;
   }
   foreach my $line (@data) {
       if($line =~ /^$key=/) {
           print $out qq|$key="$value"\n|;
           &printMsg("DEBUG","Rewrote $key = $value in $conf");
       } else {
           print $out $line;
       }
   }
   close($out);
   return 0;

}
#####################################################################
sub readLocalBoxNameserversLst() {

        my $servers_lst = shift;
        my $cmd = "cat /etc/resolv.conf | grep nameserver | /usr/bin/awk '{print \$2}'";
        my $rc = qx($cmd);

        &printMsg("INFO","Reading local box configuration.");
        $rc =~ tr/\n/ /;
        $$servers_lst = $rc;

        return 0;
}

############################################
########### PUBLIC FUNCTIONS ###############
############################################

##################################################
# getCurrentTime : Retruns curretn time stamp    #
##################################################
sub getCurrentTime() {

        my $tr = shift;
        my @td = localtime(time);
        my $ts = sprintf("%02d/%02d/%04d %02d:%02d:%02d",$td[3],$td[4]+1,$td[5]+1900,$td[2],$td[1],$td[0]);
        
        $$tr = $ts;

        return 0;
}

##################################################
#logInit : initialize logging mechanism          #
##################################################
sub logInit() {

	my $logfile = shift;

	#If logfile supplied use it instead of default
	$LOGFILE = $logfile if ( defined $logfile ); 
	$LOGINIT = 1;

    #sometimes the script gets run as root by hand and we 
    #wind up with an unwritable log file this check catches this case
    #and attempts to unlink the file before we open for write.
    if( -f $LOGFILE) {
        if ( ! -w $LOGFILE ) {
            &printMsg("WARN","Couldn't open $LOGFILE for writing.  Attempting to unlink.");
            if(unlink($LOGFILE)) {
                &printMsg("WARN","Successfully unlinked $LOGFILE");
            }
        }
    }
        eval {
            open(LFD, '>' . $LOGFILE) or die $!;
        };

	if ( $@ ) {
		&printMsg("WARN","Can not initialize logging ... Proceeding without ...");
		return 0;
	}

	return 0;
}

##################################################
#logMsg : prints info into log file              #
##################################################
sub logMsg() {
	
	my $log_type = shift;
        my $log_body = join(' ',@_);
        my $log_out;

	switch ( $log_type ) {
                case 'ERROR' { $log_out = "[ERROR] : $log_body"; }
                case 'ALERT' { $log_out = "[ALERT] : $log_body"   if ( $LOG_LEVEL >= 1 ); }
                case 'WARN'  { $log_out = "[WARN ] : $log_body"   if ( $LOG_LEVEL >= 2 ); }
                case 'INFO'  { $log_out = "[INFO ] : $log_body"   if ( $LOG_LEVEL >= 3 ); }
                case 'DISPL' { $log_out = $log_body               if ( $LOG_LEVEL >= 4 ); }
                case 'NOTE'  { $log_out = "[NOTE ] : $log_body"   if ( $LOG_LEVEL >= 5 ); }
                case 'DEBUG' { $log_out = "\t[DEBUG] : $log_body" if ( $LOG_LEVEL >= 6 ); }
        }

	if ( defined $log_out ) {
		print LFD $log_out . "\n" if ( $LOGINIT == 1 );
	}

	return 0;
}

##################################################
# printMsg : Prints/logs different level output  #
##################################################
sub printMsg() {

	my $msg_type = shift;
        my $msg_body = join(' ',@_);
        my $msg_out;	

	switch ( $msg_type ) {
		case 'ERROR' { $msg_out = "[ERROR] : $msg_body"; }
		case 'ALERT' { $msg_out = "[ALERT] : $msg_body"   if ( $DEBUG_LEVEL >= 1 ); }
		case 'WARN'  { $msg_out = "[WARN ] : $msg_body"   if ( $DEBUG_LEVEL >= 2 ); }
		case 'INFO'  { $msg_out = "[INFO ] : $msg_body"   if ( $DEBUG_LEVEL >= 3 ); }
		case 'DISPL' { $msg_out = $msg_body               if ( $DEBUG_LEVEL >= 4 ); }
		case 'NOTE'  { $msg_out = "[NOTE ] : $msg_body"   if ( $DEBUG_LEVEL >= 5 ); }
		case 'DEBUG' { $msg_out = "\t[DEBUG] : $msg_body" if ( $DEBUG_LEVEL >= 6 ); }
		else { $msg_out = "[SYS_ERROR] : Unrecognized message type [$msg_type]"; }
	}

	&logMsg("$msg_type","$msg_body") if ( $LOGINIT == 1 );

	if ( defined $msg_out ) {
		print $msg_out . "\n"; 
	}	

	return 0;
}

##################################################
# setLock : Locks running instance               #
##################################################
sub setLock() {

	&printMsg("INFO","Locking current execution instance.");

	open($FDL,">>$LOCKFILE") or die "Failed to touch lockfile [$LOCKFILE].  The following error occurred [$!]";
        unless ( flock($FDL, LOCK_EX | LOCK_NB) ) {
                &printMsg("ALERT","$0 is already running... Exiting ...");
                exit(1);
        }

	return 0;
}

##################################################
# setUnLock : Unlocks running instance           #
##################################################
sub setUnLock() {

	&printMsg("INFO","Unlocking current execution instance.");

    if(openhandle($FDL)) {
        flock($FDL, LOCK_UN);
        close($FDL);
        unlink($LOCKFILE);
    }

	return 0;
}

####################################################
# setWorkingDir : set running directory to be local#
#                 used for non local path run      #
####################################################
sub setWorkingDir() {

	my $mypath = abs_path($0);
	my $mydir = dirname($mypath);

	&printMsg("DEBUG","Switching to [$mydir] working directory.");

	eval {
		chdir $mydir or die "Failed to change working directory.";
	};

	if ( $@ ) {
		&printMsg("ERROR","$@");
		return 1;
	}
	return 0;
}

#This subroutine works very similar to alarm except it is itself safe to use 
#with code that itself calls alarm or sleep. it does this by setting up a signal handler for SIGUSR1
#and then forking a process that waits the specified amount of time before giving us the SIGUSR1 signal. 
#Since forked processes do no inherit the alarm(2) timer it is safe to use this timer without resetting
#any previously set timers in the main process or vice-versa. We also restore anything previously saved into $SIG{USR1} if anything
#when the timer is reset with a 0 interval. Optionally if a subroutine reference is passed as the second argument
#it will be used as the SIGUSR1 handler instead of the default. 
{
my $child = 0;
my $oldusr1;
    sub setAlarm {
        my $interval = shift;
        my $handler = shift;
        my $defhandler = sub { die "ThreatStop::setAlarm() Timer Expired."; };
        #Set up signal handler
        if(!$handler) {
            $handler = $defhandler;
        }
        #Save existing USR1 handler if it exists.
        #but only do this once so we don't lose the original with multiple calls to setAlarm.
        if((! $oldusr1) and $SIG{USR1}) {
            if($SIG{USR1} != $defhandler) { #Make sure we don't "restore" our own internal handler later.
                &printMsg("DEBUG","ThreatStop::setAlarm() saving existing USR1 handler for later.");
                $oldusr1 = $SIG{USR1};
            }
        }                                                                                                                                                                                       
        $SIG{USR1} = $handler;                                                                                                                                                                  
                                                                                                                                                                                                
                                                                                                                                                                                                
        if(!$interval) { #handle unsetting the alarm and signal handler.
            &printMsg("DEBUG","ThreatStop::setAlarm() unsetting alarm.");
            if($child) {
                &printMsg("DEBUG","ThreatStop::setAlarm() killing child $child");
                kill("TERM",$child);
                $child = 0;
            }
            if($oldusr1) {
                &printMsg("DEBUG","ThreatStop::setAlarm() restoring saved signal handler.");
                $SIG{USR1} = $oldusr1;
                $oldusr1 = undef;
            } else { $SIG{USR1} = undef; }
            return 0;
        } else { #Fork the simple process to signal us at a later time.
            if($child) {
                &printMsg("DEBUG","ThreatStop::setAlarm() killing child $child");
                kill("TERM",$child);
            }
            my $pid = fork();
            if($pid) {
                &printMsg("DEBUG","ThreatStop::setAlarm() forked child $pid");
                $child = $pid;
                return 0;
            } else {
                &printMsg("DEBUG","ThreatStop::setAlarm() child setting alarm to $interval seconds.");
                close($FDL) if openhandle($FDL); #alarm processes should not hold locks.
                sleep $interval;
                kill("USR1",getppid());
                exit(0);
            }
        }
}
}

##################################################
# Convert CIDR masks to dotted quads             #
##################################################
sub ipv4_cidr2msk($) {
        return join (".",  unpack 'CCCC', pack( "B*", ( "0" x shift ).( "1" x 32) ) );
}

##################################################
# os_detecter : Detects OS type                  #
##################################################
sub osPlatformDetecter() {

	my $res      = shift;
	my $res_full = shift;
	my $platform;

	$platform = qx(uname -m);
	$$res = 64 if ( $platform =~ /x86_64/ or $platform =~ /amd64/);
	$$res = 32 if ( $platform =~ /i686/   );

	chomp($platform);
	$$res_full = $platform;
	
	return 0;
}

############################################################
# osDestributionDetector : Detects OS version for Linux os #
############################################################
sub osDestributionDetector() {

    my $ver_ref = shift;

    eval {
        open (VFD, "/proc/version") or die "Cannot determine Linux Distribution\n";
    };
   
    if ( $@ ) {
        &printMsg("ERROR","$@");
        return 1;
    }

    $_ = join('',<VFD>);

    $$ver_ref = 'Unknown';
    $$ver_ref = 'RedHat' if (/Red Hat/m);
    $$ver_ref = 'Debian' if (/Debian/m);
    $$ver_ref = 'Ubuntu' if (/Ubuntu/m);

    &printMsg("DEBUG","Detected - [$$ver_ref]");

    return 0;
}

##################################################
# calculateElapsedTime :                         #
##################################################
sub calculateElapsedTime() {

	my $start_epoch      = shift;
	my $elapsed_time_ref = shift;
	my ($elapsed_time, $end_epoch);
        my ($hr, $min, $sec);

        $end_epoch = time();
        $elapsed_time = $end_epoch - $start_epoch;

        $sec = $elapsed_time % 60;
        $min = ( $elapsed_time / 60 ) % 60;
        $hr  = ( $elapsed_time / ( 60 * 60 )) % 24;

        $elapsed_time = sprintf("%02d:%02d:%02d",$hr,$min,$sec);
	$$elapsed_time_ref = $elapsed_time;
	
        return 0;
}

##################################################
# checkDNSaccess : Verifies DNS connection       #
##################################################
sub checkDNSaccess() {

	my $dig      = shift;
	my $dnslist  = shift;
	my $dnsok    = shift;
        my @dnslist  = split(/ /,$dnslist);
	my $dnsport  = shift;
        my $cmd;

        &printMsg("INFO","Checking DNS servers access ...");

	#If no dns list was received
	if ( ! @dnslist ) {
		&printMsg("ERROR","DNS servers list is not defined.");
		return 1;
	}

	if ( ! -e $dig ) {
		&printMsg('ERROR',"Missing dig [$dig] tool file.");
		return 1;
	}

	#Pass over dns servers list and try to connect
        foreach my $dns ( @dnslist ) {
	
                $cmd = "$dig -p $dnsport ptr \@$dns basic.threatstop.local 2>&1";
                &printMsg('INFO',"Verifying dns server [$dns] access.");
                &printMsg('DEBUG',"$cmd");
                my $rc = qx($cmd);
		my $rv = $?;

		if ( $rv != 0 ) {

			#Parse dig output and extract error message
			chomp($rc);
			my @rc = split('\n',$rc);
			@rc = grep {/;; (?!global options)/} @rc;
			$rc = join(' ',@rc);
			$rc =~ s/;//g;
			chomp($rc);			

			&printMsg('WARN',"dig command failed with error [$rc]");
			next;
		}

                if ( $rc =~ m/opcode: QUERY, status: NOERROR,.*QUERY: 1, ANSWER: (\d+)/s ) {
                        &printMsg('INFO',"$dns resolves OK.");
                        push(@$dnsok, $dns);
                } else {
                        &printMsg("WARN","ThreatSTOP initial dig to $dns failed, skipping to the next server in the list.");
                }
        }

        if ( ! defined @$dnsok[0] ) {
                &printMsg("ERROR","All ThreatSTOP DNS servers are unreachable.");
                return 1;
        }

        &printMsg("INFO","Done checking DNS servers.");
        return 0;
}

##########################################################
# loadConfigFileData : Load previous setup configuration #
##########################################################
sub loadConfFileData() {

	my $confile   = shift;
	my $data_hash = shift;
	my $err_msg   = shift;
	my $FILE;
    my $basedir = $FindBin::Bin;

        #Reading configuration file
        &printMsg("INFO","Initializing data from configuration file.");

        if( -e $confile ) {

            eval {
                    open (FILE, "<", $confile) or die "$!";
            };
        } else {
            eval {
                    open (FILE, "<", "$basedir/$confile") or die "$!";
                    $confile = "$basedir/$confile";
            };
        }

        if ( $@ ) {
                $$err_msg = "$@";
                return 1;
        }

        #Parse data
        while (<FILE>) {
                next unless (/^([a-z]\w+)=[("]?(.*?)[")]?\s+$/);
                $data_hash->{$1}=$2;
        }
        close FILE;
        mpw($confile,$data_hash);
	#Also update DNS servers list with pre-stored setting
	$DNS_LIST = $dns_server if ( defined $dns_server );

        return 0;
}

############################################################
# appendUploadLogEntry : Append an entry to the Upload Log #
############################################################
sub appendUploadLogEntry() {

	my $orig_filename    = shift;
	my $chunked_filename = shift;
	my $status_code      = shift;
	my $transfer_time    = shift;

	my $status           = "";
	my $status_msg       = "";
	chomp(my $timestamp = `date -Iseconds`);
	my $FILE;
	
	switch ( $status_code ) {
		case 0 { $status = "SUCCESS"; $status_msg = "Transfer time (seconds): ${transfer_time}"}
		case 1 { $status = "FAIL"; $status_msg = "Connection failed"; }
		case 2 { $status = "FAIL"; $status_msg = "Connection established, but send status was not OK"; }
		else { $status = "UNKNOWN"; $status_msg = "Unrecognized status code [$status_code]"; }
	}

	open (FILE, ">>", "${UPLOADSLOGFILE}${curr_date}.log") or die "$!";
	print FILE "$timestamp   $orig_filename   $chunked_filename   $status   $status_msg\n";
	close FILE;

	return 0;
}

##################################################################
# purgeOldUploadsLogs : Removes logs that are over ~2 months old #
#
# NOTE:  It would probably be better to implement this as a cron
#        instead of in ThreatStop.pm.  The setup routine could
#        be modified to accomplish this task.
##################################################################
sub purgeOldUploadsLogs() {

	my $logsdir = $FindBin::Bin;

	my $rv = `find $logsdir -name 'tsuploads??????.log' -mtime +63 -exec rm -f {} \\;`;

	return 0;
}

##################################################################
# querySRV
# 
# Using the passed service, protocol and domain, this will
# load the passed array reference as follows :
#
# args :
#   service name (without the underscore)
#   protocol name (without the underscore)
#   domain name to which the query should be made
#   reference to an array, which will get loaded as follows :
#
# $ref = [
#          { # record 0
#              priority => ...,
#              weight => ...,
#              target => ...,
#              port => ...,
#          },
#          { # record 1
#          ...
#          },
#        ]
#
# returns :
# 1 for success
# 0 for failure, which could mean bad input, or unable to query,
#   or empty return
#
sub querySRV {

    my $dig      = shift;
    my $service  = shift;
    my $protocol = shift;
    my $domain   = shift;
    my $arrayref = shift;

    &printMsg("DEBUG", "SRV query for '_$service._$protocol.$domain'");

    my $srvs = `$dig +tcp +short -t srv _$service._$protocol.$domain 2>&1`;

    my $status = $?;
  
    &printMsg("DEBUG", "SRV query status ($status) return : $srvs");
    if ( $status != 0 ) {

        return 0;
    }

    my ($priority, $weight, $port, $target) = ();
    foreach my $srv ( split(/\n/, $srvs) ) {

        &printMsg("DEBUG", "working on line : '$srv'");

        if ( $srv =~ /^\s+$/ ) {

            &printMsg("DEBUG", "Got an empty SRV '$srv'; skipping");
            next;
        }
 
        ($priority, $weight, $port, $target) = split(/\s+/, $srv);

        &printMsg("DEBUG", "SRV : $priority, $weight, $target, $port");
        push @$arrayref, {
                          priority => $priority,
                          service => $service,
                          port => $port,
                          target => $target,
                         };
    }

    return 0 if ! scalar @$arrayref;
    return 1;
}

###############################################################
##################### MODULE END ##############################
###############################################################

1;
